#!/bin/bash

SCRIPT_DIR=$(realpath "$(dirname "$0")")
. "$SCRIPT_DIR/secureboot.conf"

# constants
IMXBOOT_LOAD_ADDR=0x40480000
IMXBOOT_DEKBLOB_ADDR=0x40400000
LINUX_LOAD_ADDR=0x80080000
LINUX_FDT_ADDR=0x45000000
LINUX_FDT_OVERLAY_ADDR=0x45020000

CSF_HEADER="[Header]
Version = 4.3
Hash Algorithm = sha256
Engine = CAAM
Engine Configuration = 0
Certificate Format = X509
Signature Format = CMS

[Install SRK]
File = \"$CST/crts/SRK_1_2_3_4_table.bin\"
Source index = ${CST_SOURCE_INDEX}

[Install CSFK]
File = \"$CST/crts/CSF$((CST_SOURCE_INDEX+1))_1_sha256_${CST_KEYTYPE}_v3_usr_crt.pem\"

[Authenticate CSF]

[Install Key]
Verification index = 0
Target index = 2
File = \"$CST/crts/IMG$((CST_SOURCE_INDEX+1))_1_sha256_${CST_KEYTYPE}_v3_usr_crt.pem\""

case "$CST_UNLOCK_SRK" in
[1yY]*)
	CSF_HEADER="$CSF_HEADER

[Unlock]
Engine = OCOTP
Features = SRK REVOKE"
	;;
esac


error() {
	printf "%s\n" "$@"
	exit 1
}

check_cst_srk() {
        [[ -e "$CST/crts/SRK_1_2_3_4_table.bin" ]] \
                || error "SRK not setup yet, please run secureboot.sh init_srk first"
        [[ -e "$CST/keys/SRK$((CST_SOURCE_INDEX+1))_sha256_${CST_KEYTYPE}_v3_ca_key.pem" ]] \
                || error "missing SRK key $CST/keys/SRK$((CST_SOURCE_INDEX+1))_sha256_${CST_KEYTYPE}_v3_ca_key.pem"
        [[ -e "$CST/keys/IMG1_1_sha256_${CST_KEYTYPE}_v3_usr_key.pem" ]] \
                || error "SRK was not init for habv4 (Armadillo IoT G4/X2's imx8mp), extract directory anew and rerun secureboot.sh init_srk"
}


pre_checks() {
	local overlay
	mkdir -p "$OUT"
	[[ -e "$CST" ]] || error "Please download and extract https://www.nxp.com/search?keyword=IMX_CST_TOOL_NEW (NXP account required)"
	case "$1" in
	imxboot*)
		check_cst_srk
		[[ -e "$IMXBOOT" ]] || error "Please download and extract imx-boot"
		;;
	linux*)
		check_cst_srk
		[[ -e "$LINUX_IMAGE" ]] || error "Please build or extract linux from rootfs ($LINUX_IMAGE not found)"
		[[ -e "$LINUX_DTB" ]] || error "$LINUX_DTB not found"
		for overlay in "${LINUX_DTB_OVERLAYS[@]}"; do
			[[ -e "${LINUX_DTB%/*}/$overlay" ]] \
				|| error "$overlay not found in ${LINUX_DTB%/*}"
		done
		[[ -z "$LINUX_INITRD" ]] || [[ -e "$LINUX_INITRD" ]] \
			|| error "$LINUX_INITRD not found"
		;;
	esac
}

init_srk() {
	# already done
	[[ -e "$CST/crts/SRK_1_2_3_4_table.bin" ]] \
		&& error "SRK has already been setup, refusing to run"

	(
		cd "$CST/keys" || exit 1
		echo 00 > serial || error "Could not write serial"
		if ! [[ -e "key_pass.txt" ]]; then
			echo "Generating random passphrase for CST keys (written as plain text in $CST/keys/key_pass.txt)"
			KEYPASS=$(tr -cd '0-9a-zA-Z#@!.[]_-' < /dev/urandom  | head -c 20)
			printf "%s\n" "$KEYPASS" "$KEYPASS" > key_pass.txt \
				|| error "Could not write key_pass.txt"
		fi
		# new 3.3.2 syntax
		ARGS=( -existing-ca n -num-srk 4 -srk-ca y )
		if grep -q -- '-kt' hab4_pki_tree.sh; then
			case "$CST_ECC" in
			y) ARGS+=( -kt ecc );;
			*) ARGS+=( -kt rsa );;
			esac
		else
			ARGS+=( -use-ecc "$CST_ECC" )
		fi
		ARGS+=( -kl "$CST_KEYLEN" -duration "$CST_DURATION" )
		./hab4_pki_tree.sh "${ARGS[@]}"
	) || error "Could not generate hab4 pki tree"
	(
		cd "$CST/crts" || exit 1
		../linux64/bin/srktool --hab_ver 4 --digest sha256 \
		        --table SRK_1_2_3_4_table.bin --efuses SRK_1_2_3_4_fuse.bin \
			--fuse_format 1 \
			--certs "./SRK1_sha256_${CST_KEYTYPE}_v3_ca_crt.pem,./SRK2_sha256_${CST_KEYTYPE}_v3_ca_crt.pem,./SRK3_sha256_${CST_KEYTYPE}_v3_ca_crt.pem,./SRK4_sha256_${CST_KEYTYPE}_v3_ca_crt.pem"
	) || error "Could not generate SRK table/fuse map"

	echo
	echo "SRK has been initialized. Backup $CST to a secure location !!"
}

print_srk_fuse() {
	local -a FUSES
	mapfile -t FUSES < <(hexdump -e '1/4 "0x"' -e '1/4 "%X""\n"' "$CST/crts/SRK_1_2_3_4_fuse.bin")
	local IFS=" "
	echo
	echo "Use the following commands in uboot to set root key hash in OTP:"
	echo "fuse prog -y 6 0 ${FUSES[*]:0:4}"
	echo "fuse prog -y 7 0 ${FUSES[*]:4:4}"
	echo "fuse prog -y 0 0 0x200"
	echo
	echo
	echo "For testing it is possible to temporarly override fuses instead,"
	echo "but this only validates the last part of the boot:"
	echo "fuse override 6 0 ${FUSES[*]:0:4}"
	echo "fuse override 7 0 ${FUSES[*]:4:4}"

}

# imxboot helper: get offset from build log
get_offset() {
	local key="$1"
	local target="$2"

	awk "/OFFSET dump/ { dump=1 }
	     dump && /make/ { dump=0 }
	     dump && /$key/ { print $target }" "$OUT/${IMXBOOT_TARGET}.buildlog"
}

#########################################################################
# normal secure boot, imxboot image                                     #
#                                                                       #
# build imxboot image and sign it                                       #
#########################################################################

imxboot() {
	mkdir -p "$OUT"
	make firmware \
		|| error "Could not get firmwares"
	make -j5 -C "$IMXBOOT" "imx-boot_${IMXBOOT_TARGET}" 2>&1 \
			| tee "$OUT/${IMXBOOT_TARGET}.buildlog" \
		|| error "Build of imxboot failed"

	local block
	# shellcheck disable=SC2016 ## single quotes are on purpose
	block=$(get_offset "spl hab block" '$4, $5, $6')
	[[ -n "$block" ]] || error "Could not get spl hab block from build logs"
	cat > "$OUT/${IMXBOOT_TARGET}.csf_spl.txt" <<EOF \
		|| error "Could not write $OUT/${IMXBOOT_TARGET}.csf_spl.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks =  $block "$IMXBOOT/imx-boot_${IMXBOOT_TARGET}"

[Unlock]
Engine = CAAM
Features = MID
EOF
	{
		cat <<EOF
$CSF_HEADER

[Authenticate Data]
Verification index = 2
EOF
		# shellcheck disable=SC2016 ## single quotes are on purpose
		block=$(get_offset "sld hab block" '$4, $5, $6')
		[[ -n "$block" ]] || error "Could not get sld hab block from build logs"
		echo -n "Blocks = $block \"$IMXBOOT/imx-boot_${IMXBOOT_TARGET}\"";
		while read -r block; do
			echo ", \\"
			echo -n "$block \"$IMXBOOT/imx-boot_${IMXBOOT_TARGET}\""
		done < <(awk '/\.\/print_fit_hab\.sh/ { dump=1; getline }
			      dump && /make/ { dump=0 }
			      dump' "$OUT/${IMXBOOT_TARGET}.buildlog")
		echo
	} > "$OUT/${IMXBOOT_TARGET}.csf_fit.txt" \
		|| error "Could not write $OUT/${IMXBOOT_TARGET}.csf_fit.txt"

	"$CST/linux64/bin/cst" -i "$OUT/${IMXBOOT_TARGET}.csf_spl.txt" -o "$OUT/${IMXBOOT_TARGET}.csf_spl.bin" \
		|| error "Could not sign csf spl"
	"$CST/linux64/bin/cst" -i "$OUT/${IMXBOOT_TARGET}.csf_fit.txt" -o "$OUT/${IMXBOOT_TARGET}.csf_fit.bin" \
		|| error "Could not sign csf fit"
	cp "$IMXBOOT/imx-boot_${IMXBOOT_TARGET}" "$OUT/imx-boot_${IMXBOOT_TARGET}.signed"
	local offset
	# shellcheck disable=SC2016 ## single quotes are on purpose
	offset=$(get_offset "^ *csf_off" '$2')
	[[ -n "$offset" ]] || error "Could not get csf_off from build logs"
	dd if="$OUT/${IMXBOOT_TARGET}.csf_spl.bin" of="$OUT/imx-boot_${IMXBOOT_TARGET}.signed" \
		seek=$((offset)) oflag=seek_bytes bs=4K conv=notrunc

	# shellcheck disable=SC2016 ## single quotes are on purpose
	offset=$(get_offset "^ sld_csf_off" '$2')
	[[ -n "$offset" ]] || error "Could not get sld_csf_off from build logs"
	dd if="$OUT/${IMXBOOT_TARGET}.csf_fit.bin" of="$OUT/imx-boot_${IMXBOOT_TARGET}.signed" \
		seek=$((offset)) oflag=seek_bytes bs=4K conv=notrunc

	echo "Created $OUT/imx-boot_${IMXBOOT_TARGET}.signed"
}

#########################################################################
# normal secure boot, linux image                                       #
#                                                                       #
# build linux fit image from configured settings and sign it            #
#########################################################################

maybe_compress() {
	local file
	for file; do
		[[ -e "$file" ]] || continue
		[[ -s "$file.zst" ]] && [[ "$file.zst" -nt "$file" ]] && continue
		zstd -c "$file" > "$file.zst" \
			|| error "Could not compress $file"
	done
}

# helper: build fit image from linux image and dtbs
linux_build_fit() {
	local overlay

	[ -e "$IMXBOOT/uboot-imx/$IMXBOOT_TARGET/tools/mkimage" ] \
		|| error "mkimage not found, build imxboot first"

	# Note: each entry could have a 'hash-1 { algo = "sha1"; };' block, but
	# since we sign content it is redundant.
	# The hash might also not be compatible with encryption as we might
	# only encrypt the content and verification would happen too early
	{
		cat <<EOF
/dts-v1/;
/ {
	description = "fit linux image for armadillo";
	#address-cells = <1>;
	images {
		kernel {
			description = "linux kernel";
			data = /incbin/("$LINUX_IMAGE.zst");
			type = "kernel";
			arch = "arm64";
			os = "linux";
			compression = "zstd";
			load = <$LINUX_LOAD_ADDR>;
			entry = <$LINUX_LOAD_ADDR>;
		};
		fdt {
			description = "fdt";
			data = /incbin/("$LINUX_DTB.zst");
			type = "flat_dt";
			arch = "arm64";
			compression = "zstd";
			load = <$LINUX_FDT_ADDR>;
		};
EOF
		if [[ -n "$LINUX_INITRD" ]]; then
			cat <<EOF
		ramdisk {
			description = "initrd";
			data = /incbin/("$LINUX_INITRD");
			type = "ramdisk";
			arch = "arm64";
			os = "linux";
			compression = "none";
		};
EOF
		fi
		for overlay in "${LINUX_DTB_OVERLAYS[@]}"; do
			fullpath="${LINUX_DTB%/*}/$overlay"
			cat <<EOF
		$overlay {
			description = "$overlay";
			data = /incbin/("$fullpath.zst");
			type = "flat_dt";
			arch = "arm64";
			compression = "zstd";
			load = <$LINUX_FDT_OVERLAY_ADDR>;
		};
EOF
		done
		cat <<EOF
	};
	configurations {
		default = "armadillo";
		armadillo {
			description = "armadillo";
			kernel = "kernel";
			fdt = "fdt";
EOF
		if [[ -n "$LINUX_INITRD" ]]; then
			cat <<EOF
			ramdisk = "ramdisk";
EOF
		fi
			cat <<EOF
		};
EOF
		for overlay in "${LINUX_DTB_OVERLAYS[@]}"; do
			cat <<EOF
		$overlay {
			fdt = "$overlay";
		};
EOF
		done
		cat <<EOF
	};
};
EOF
	} > "$OUT/Image.its" \
		|| error "Could not write linux fit configuration"

	maybe_compress "$LINUX_IMAGE" "$LINUX_DTB"
	for overlay in "${LINUX_DTB_OVERLAYS[@]}"; do
		maybe_compress "${LINUX_DTB%/*}/$overlay"
	done

	"$IMXBOOT/uboot-imx/$IMXBOOT_TARGET/tools/mkimage" -f "$OUT/Image.its" "$OUT/Image.fit" \
		|| error "Could not build linux fit image"
}


linux() {
	local linux_size
	linux_build_fit
	linux_size=$(stat -c "%s" "$OUT/Image.fit")
	[[ -n "$linux_size" ]] || error "Could not get linux image size"
	# round up to 0x1000 as pert uboot authenticate_image()
	linux_size=$((((linux_size + 0xfff) / 0x1000) * 0x1000))
	mv "$OUT/Image.fit" "$OUT/Image_signing.bin" \
		|| error "Could not copy FIT image"
	cat > "$OUT/genIVT.pl" <<EOF
#!/usr/bin/perl -w
use strict;
print pack("V", 0x412000D1); # Signature
print pack("V", ${IMXBOOT_LOAD_ADDR}); # Load Address (*load_address)
print pack("V", 0x0); # Reserved
print pack("V", 0x0); # DCD pointer
print pack("V", 0x0); # Boot Data
print pack("V", $(printf "%#x" $((IMXBOOT_LOAD_ADDR + linux_size)))); # Self Pointer (*ivt)
print pack("V", $(printf "%#x" $((IMXBOOT_LOAD_ADDR + linux_size + 0x20)))); # CSF Pointer (*csf)
print pack("V", 0x0); # Reserved
EOF
	perl "$OUT/genIVT.pl" \
			| dd of="$OUT/Image_signing.bin" bs=1k status=none \
				seek=$linux_size oflag=seek_bytes \
		|| error "Could not generate linux IVT"

	cat > "$OUT/Image.csf_linux.txt" <<EOF \
		|| error "Could not write $OUT/Image.csf_linux.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks = $IMXBOOT_LOAD_ADDR 0x00000000 $(printf "%#x" $((linux_size + 0x20))) "$OUT/Image_signing.bin"
EOF
	"$CST/linux64/bin/cst" -i "$OUT/Image.csf_linux.txt" -o "$OUT/Image.csf_linux.bin" \
		|| error "Could not sign csf linux"
	cat "$OUT/Image.csf_linux.bin" >> "$OUT/Image_signing.bin" \
		|| error "Could not append to $OUT/Image_signing.bin"
	chmod 644 "$OUT/Image_signing.bin" \
		|| error "Could not fix image permission"
	mv "$OUT/Image_signing.bin" "$OUT/Image.signed" \
		|| error "Could not rename to $OUT/Image.signed"
	echo "Created $OUT/Image.signed"
}


#########################################################################
# encrypted secure boot, imxboot image                                  #
#                                                                       #
# build imxboot image, encrypt it in multiple steps and sign the result #
# 'dek key' is included as is in encrypted image and must be processed  #
# on each device before use                                             #
#########################################################################

imxboot_enc() {
	# sanity check until bug with 3.3.2 is fixed
	if ! "$CST/linux64/bin/cst" --version | grep -q 3.3.1; then
		error "imxboot encryption is only supported with cst 3.3.1, please downgrade"
	fi

	mkdir -p "$OUT"

	## rebuild imx-boot image to get build logs
	# for encryption we need dek_blob_fit_dummy.bin to be present to reserve
	# space in fit image for final dek blob
	mkdir -p "$IMXBOOT/imx-mkimage/${IMXBOOT_TARGET}"
	dd if=/dev/zero of="$IMXBOOT/imx-mkimage/${IMXBOOT_TARGET}/dek_blob_fit_dummy.bin" \
			bs=96 count=1 status=none \
		|| error "Could not create dummy dek blob for fit image"
	make -j5 -C "$IMXBOOT" "imx-boot_${IMXBOOT_TARGET}" 2>&1 \
			| tee "$OUT/${IMXBOOT_TARGET}.buildlog" \
		|| error "Build of imxboot failed"
	# remove dummy dek blob as it would affect normal builds that could follow
	rm -f "$IMXBOOT/imx-mkimage/${IMXBOOT_TARGET}/dek_blob_fit_dummy.bin"


	## generate csf configuration files for SPL uboot
	# note we cannot both sign and encrypt in a single step so we need
	# to generate two csf files
	local block
	# shellcheck disable=SC2016 ## single quotes are on purpose
	block=$(get_offset "spl hab block" '$4, $5, $6')
	[[ -n "$block" ]] || error "Could not get spl hab block from build logs"
	local block_start=${block%% *}
	local block_length=${block##* }
	local blob_address
	# Authenticate Start Address +  SPL & DDR FW image length + CSF Padding
	blob_address=$(printf "%#x" $((block_start + block_length + 0x2000)))
	local decrypt_block
	# Start Address = Start Address + SPL header = 0x91ffc0 + 0x40 = 0x920000
	# Offset = Image offset (image_off) = 0x40
	# Decrypt size = Image length - SPL header = 0x24a00 - 0x40 = 0x249c0
	decrypt_block=$(printf "%#x 0x40 %#x" \
			$((block_start + 0x40)) \
			$((block_length - 0x40)))
	cat > "$OUT/${IMXBOOT_TARGET}.csf_spl_enc.txt" <<EOF \
		|| error "Could not write $OUT/${IMXBOOT_TARGET}.csf_spl_enc.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks =  $block_start 0x0 0x40 "$IMXBOOT/imx-boot_${IMXBOOT_TARGET}"

[Unlock]
Engine = CAAM
Features = MID

[Install Secret Key]
    Verification Index = 0
    Target Index = 0
    Key = "$OUT/${IMXBOOT_TARGET}.dek_spl.bin"
    Key Length = 256
    # Authenticate Start Address +  SPL & DDR FW image length + CSF Padding
    Blob Address = $blob_address

[Decrypt Data]
    # The decrypt data command below causes CST to modify the input
    # file and encrypt the specified block of data. This image file
    # is a copy of the file used for the authentication command above
    Verification Index = 0
    Mac Bytes = 16
    # Start Address = Start Address + SPL header
    # Offset = Image offset (image_off) = 0x40
    # Decrypt size = Image length - SPL header
    Blocks = $decrypt_block "$OUT/imx-boot_${IMXBOOT_TARGET}.enc"
EOF
	cat > "$OUT/${IMXBOOT_TARGET}.csf_spl_sign_enc.txt" <<EOF \
		|| error "Could not write $OUT/${IMXBOOT_TARGET}.csf_spl_sign_enc.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks =  $block "$OUT/imx-boot_${IMXBOOT_TARGET}.enc"

[Unlock]
Engine = CAAM
Features = MID

[Install Secret Key]
    Verification Index = 0
    Target Index = 0
    Key = "$OUT/${IMXBOOT_TARGET}.dek_spl_dummy.bin"
    Key Length = 256
    # Authenticate Start Address +  SPL & DDR FW image length + CSF Padding
    Blob Address = $blob_address

[Decrypt Data]
    # The decrypt data command below causes CST to modify the input
    # file and encrypt the specified block of data. This image file
    # is a copy of the file used for the authentication command above
    Verification Index = 0
    Mac Bytes = 16
    # Start Address = Start Address + SPL header
    # Offset = Image offset (image_off) = 0x40
    # Decrypt size = Image length - SPL header
    Blocks = $decrypt_block "$OUT/imx-boot_${IMXBOOT_TARGET}.enc.dummy"
EOF


	## generate csf configuration file for FIT image (2nd stage:
	# uboot, optee, atf)

	# shellcheck disable=SC2016 ## single quotes are on purpose
	block=$(get_offset "sld hab block" '$4, $5, $6')
	[[ -n "$block" ]] || error "Could not get sld hab block from build logs"
	local fit_line1 fit_line2 fit_line3 fit_line4
	{
		read -r fit_line1
		read -r fit_line2
		read -r fit_line3
		read -r fit_line4
	} < <(awk '/\.\/print_fit_hab\.sh/ { dump=1; getline }
		   dump && /make/ { dump=0 }
		   dump' "$OUT/${IMXBOOT_TARGET}.buildlog")
	[[ -n "$fit_line1" ]] && [[ -n "$fit_line2" ]] \
		&& [[ -n "$fit_line3" ]] && [[ -n "$fit_line4" ]] \
		|| error "print_fit_hab format not as expected"
	# blocks need to be 16bit aligned, but first two lines are not
	# (and happen to be contiguous): merge these. Geez.
	local fit_line1_start fit_line1_offset fit_line1_len
	fit_line1_start=${fit_line1%% *}
	fit_line1_offset=${fit_line1% *}
	fit_line1_offset=${fit_line1_offset##* }
	fit_line1_len=${fit_line1##* }
	local fit_line2_start fit_line2_offset fit_line2_len
	fit_line2_start=${fit_line2%% *}
	fit_line2_offset=${fit_line2% *}
	fit_line2_offset=${fit_line2_offset##* }
	fit_line2_len=${fit_line2##* }
	# check 2nd line start/offset are 1st line + length
	(( fit_line2_start == fit_line1_start + fit_line1_len )) \
		&& (( fit_line2_offset == fit_line1_offset + fit_line1_len )) \
		|| error "print_fit_hab first two lines are not contiguous"
	fit_line1=$(printf "%s %s %#x" "$fit_line1_start" "$fit_line1_offset" \
			"$((fit_line1_len + fit_line2_len))")

	cat > "$OUT/${IMXBOOT_TARGET}.csf_fit_enc.txt" <<EOF \
		|| error "Could not write $OUT/${IMXBOOT_TARGET}.csf_fit_enc.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks = $block "$OUT/imx-boot_${IMXBOOT_TARGET}.enc"

[Install Secret Key]
    Verification Index = 0
    Target Index = 0
    Key = "$OUT/${IMXBOOT_TARGET}.dek_fit.bin"
    Key Length = 256
    Blob Address = ${IMXBOOT_DEKBLOB_ADDR}

[Decrypt Data]
    Verification Index = 0
    Mac Bytes = 16
    Blocks = $fit_line1 "$OUT/imx-boot_${IMXBOOT_TARGET}.enc", \\
             $fit_line3 "$OUT/imx-boot_${IMXBOOT_TARGET}.enc", \\
             $fit_line4 "$OUT/imx-boot_${IMXBOOT_TARGET}.enc"
EOF
	cat > "$OUT/${IMXBOOT_TARGET}.csf_fit_sign_enc.txt" <<EOF \
		|| error "Could not write $OUT/${IMXBOOT_TARGET}.csf_fit_sign_enc.txt"
$CSF_HEADER

[Authenticate Data]
Verification index = 2
Blocks = $block "$OUT/imx-boot_${IMXBOOT_TARGET}.enc", \\
         $fit_line1 "$OUT/imx-boot_${IMXBOOT_TARGET}.enc", \\
         $fit_line3 "$OUT/imx-boot_${IMXBOOT_TARGET}.enc", \\
         $fit_line4 "$OUT/imx-boot_${IMXBOOT_TARGET}.enc"

[Install Secret Key]
    Verification Index = 0
    Target Index = 0
    Key = "$OUT/${IMXBOOT_TARGET}.dek_fit_dummy.bin"
    Key Length = 256
    Blob Address = ${IMXBOOT_DEKBLOB_ADDR}

[Decrypt Data]
    Verification Index = 0
    Mac Bytes = 16
    Blocks = $fit_line1 "$OUT/imx-boot_${IMXBOOT_TARGET}.enc.dummy", \\
             $fit_line3 "$OUT/imx-boot_${IMXBOOT_TARGET}.enc.dummy", \\
             $fit_line4 "$OUT/imx-boot_${IMXBOOT_TARGET}.enc.dummy"
EOF


	## encrypt then sign the SPL image

	cp "$IMXBOOT/imx-boot_${IMXBOOT_TARGET}" "$OUT/imx-boot_${IMXBOOT_TARGET}.enc" \
		|| error "Could not copy imx-boot_${IMXBOOT_TARGET}"

	"$CST/linux64/bin/cst" -i "$OUT/${IMXBOOT_TARGET}.csf_spl_enc.txt" -o "$OUT/${IMXBOOT_TARGET}.csf_spl_enc.bin" \
		|| error "Could not encrypt csf spl"

	cp "$OUT/imx-boot_${IMXBOOT_TARGET}.enc" "$OUT/imx-boot_${IMXBOOT_TARGET}.enc.dummy" \
		|| error "Could not copy imx-boot_${IMXBOOT_TARGET}.enc.dummy"
	"$CST/linux64/bin/cst" -i "$OUT/${IMXBOOT_TARGET}.csf_spl_sign_enc.txt" -o "$OUT/${IMXBOOT_TARGET}.csf_spl_sign_enc.bin" \
		|| error "Could not sign encrypted csf spl"
	rm -f "$OUT/imx-boot_${IMXBOOT_TARGET}.enc.dummy" "$OUT/${IMXBOOT_TARGET}.dek_spl_dummy.bin"


	## copy encryption nounce from first csf output into final signed one

	local offset_in offset_out
	offset_in=$(stat -c "%s" "$OUT/${IMXBOOT_TARGET}.csf_spl_enc.bin")
	[[ -n "$offset_in" ]] || error "Could not get csf_spl_enc.bin size"
	offset_in=$((offset_in - 36))
	offset_out=$(stat -c "%s" "$OUT/${IMXBOOT_TARGET}.csf_spl_sign_enc.bin")
	[[ -n "$offset_out" ]] || error "Could not get csf_spl_sign_enc.bin size"
	offset_out=$((offset_out - 36))
	echo "copying csf_spl_enc.bin $offset_in -> csf_spl_sign_enc.bin $offset_out 36 bytes"
	dd if="$OUT/${IMXBOOT_TARGET}.csf_spl_enc.bin" \
			of="$OUT/${IMXBOOT_TARGET}.csf_spl_sign_enc.bin" \
			status=none bs=36 count=1 \
			iflag=skip_bytes oflag=seek_bytes \
			skip="$offset_in" seek="$offset_out" \
		|| error "Could not copy nonce/mac to csf_spl_sign_enc"

	## encrypt then sign the FIT image

	"$CST/linux64/bin/cst" -i "$OUT/${IMXBOOT_TARGET}.csf_fit_enc.txt" -o "$OUT/${IMXBOOT_TARGET}.csf_fit_enc.bin" \
		|| error "Could not encrypt csf fit"
	cp "$OUT/imx-boot_${IMXBOOT_TARGET}.enc" "$OUT/imx-boot_${IMXBOOT_TARGET}.enc.dummy" \
		|| error "Could not copy imx-boot_${IMXBOOT_TARGET}.enc.dummy"
	"$CST/linux64/bin/cst" -i "$OUT/${IMXBOOT_TARGET}.csf_fit_sign_enc.txt" -o "$OUT/${IMXBOOT_TARGET}.csf_fit_sign_enc.bin" \
		|| error "Could not sign encrypted csf fit"
	rm -f "$OUT/imx-boot_${IMXBOOT_TARGET}.enc.dummy" "$OUT/${IMXBOOT_TARGET}.dek_fit_dummy.bin"

	## copy encryption nounce from first csf output into final signed one

	offset_in=$(stat -c "%s" "$OUT/${IMXBOOT_TARGET}.csf_fit_enc.bin")
	[[ -n "$offset_in" ]] || error "Could not get csf_fit_enc.bin size"
	offset_in=$((offset_in - 36))
	offset_out=$(stat -c "%s" "$OUT/${IMXBOOT_TARGET}.csf_fit_sign_enc.bin")
	[[ -n "$offset_out" ]] || error "Could not get csf_fit_sign_enc.bin size"
	offset_out=$((offset_out - 36))
	echo "copying csf_fit_enc.bin $offset_in -> csf_fit_sign_enc.bin $offset_out 36 bytes"
	dd if="$OUT/${IMXBOOT_TARGET}.csf_fit_enc.bin" \
			of="$OUT/${IMXBOOT_TARGET}.csf_fit_sign_enc.bin" \
			status=none bs=36 count=1 \
			iflag=skip_bytes oflag=seek_bytes \
			skip="$offset_in" seek="$offset_out" \
		|| error "Could not copy nonce/mac to csf_fit_sign_enc"


	# write final spl csf output into boot image

	local offset
	# shellcheck disable=SC2016 ## single quotes are on purpose
	offset=$(get_offset "^ csf_off" '$2')
	[[ -n "$offset" ]] || error "Could not get csf_off from build logs"
	: "copying csf_spl_sign_enc.bin (full) -> final boot image $offset"
	dd if="$OUT/${IMXBOOT_TARGET}.csf_spl_sign_enc.bin" \
		of="$OUT/imx-boot_${IMXBOOT_TARGET}.enc" \
		status=none seek=$((offset)) oflag=seek_bytes bs=4K conv=notrunc


	## write spl dek raw key into boot image
	# note we write dek_spl here but it is not final:
	# it needs to be encapsulated for the device on the device
	# we nevertheless use the same emplacement in the file for convenience
	# as encapsulated version is bigger and will overwrite this fully
	# dek offsets are stored in dek_offsets file for swupdate/manual flashing


	offset=$((offset + 0x2000))
	printf "offset for dek_spl_blob: %#x\n" "$offset"
	printf "dek_spl_offset %#x " "$offset" > "$OUT/${IMXBOOT_TARGET}.dek_offsets"
	dd if="$OUT/${IMXBOOT_TARGET}.dek_spl.bin" \
		of="$OUT/imx-boot_${IMXBOOT_TARGET}.enc" \
		status=none seek="$offset" oflag=seek_bytes bs=4K conv=notrunc


	## same with fit csf binary and dek

	# shellcheck disable=SC2016 ## single quotes are on purpose
	offset=$(get_offset "^ sld_csf_off" '$2')
	[[ -n "$offset" ]] || error "Could not get sld_csf_off from build logs"
	: "copying csf_fit_sign_enc.bin (full) -> final boot image $offset"
	dd if="$OUT/${IMXBOOT_TARGET}.csf_fit_sign_enc.bin" \
		of="$OUT/imx-boot_${IMXBOOT_TARGET}.enc" \
		status=none seek=$((offset)) oflag=seek_bytes bs=4K conv=notrunc

	offset=${fit_line4#* }
	offset=$((${offset#* } + ${offset% *}))
	printf "offset for dek_fit_blob: %#x\n" "$offset"
	printf "dek_fit_offset %#x\n" "$offset" >> "$OUT/${IMXBOOT_TARGET}.dek_offsets"
	dd if="$OUT/${IMXBOOT_TARGET}.dek_fit.bin" \
		of="$OUT/imx-boot_${IMXBOOT_TARGET}.enc" \
		status=none seek="$offset" oflag=seek_bytes bs=4K conv=notrunc

	echo "Created $OUT/imx-boot_${IMXBOOT_TARGET}.enc"
}

help() {
	echo "Usage: $0 action"
	echo
	echo "Where action is one of:"
	echo "init_srk: initialize super root key"
	echo "print_srk_fuse: Show uboot commands to burn fuses for generated key"
	echo "imxboot: build and sign imx-boot image"
	echo "imxboot_enc: build, encrypt and sign imx-boot image."
	echo "             Note this requires special handling to install."
	echo "linux: build a fit image from configured linux files and sign it"
	echo
	echo "See $SCRIPT_DIR/secureboot.conf for configuration settings"
}

main() {
	local arg

	(( $# == 0 )) && help

	for arg; do
		case "$arg" in
		imxboot|imxboot_enc|linux|init_srk|print_srk_fuse) ;;
		help|-h|--help) help; exit 0;;
		*) echo "Unrecognized option $arg"; echo; help; exit 1;;
		esac

		( pre_checks "$arg"; ) \
			|| error "pre_checks failed, if the error is not obvious please check paths in secureboot.conf"
	done

	for arg; do
		"$arg"
	done
}

main "$@"
